#!/usr/bin/python

import stat
import sys

from portage import os
from portage.checksum import perform_md5
from portage.util import writemsg, writemsg_stdout

def make_hardlinks(dir1, dir2):
	"""
	Atomically create hardlinks from dir1 to dir2, for files
	with identical size and MD5 sum. Efficiently detects existing
	hardlinks by comparing inode numbers. Uses portage API for
	safe handling of UTF-8 file names, regardless of locale.
	"""
	try:
		set1 = set(os.listdir(dir1))
	except OSError:
		usage()
		return 1

	try:
		set2 = set(os.listdir(dir2))
	except OSError:
		usage()
		return 1

	pid = os.getpid()

	for filename in set1.intersection(set2):
		path1 = os.path.join(dir1, filename)
		path2 = os.path.join(dir2, filename)
		st1 = os.lstat(path1)
		st2 = os.lstat(path2)
		if not (stat.S_ISREG(st1.st_mode) and stat.S_ISREG(st2.st_mode)):
			continue
		if st1.st_size != st2.st_size:
			continue
		if (st1.st_dev, st1.st_ino) == (st2.st_dev, st2.st_ino):
			continue
		if perform_md5(path1)[0] != perform_md5(path2)[0]:
			continue
		if st1.st_nlink >= st2.st_nlink:
			p1 = path1
			p2 = path2
		else:
			p1 = path2
			p2 = path1
		tmp2 = '%s.%s' % (path2, pid)
		try:
			os.link(p1, tmp2)
		except OSError as e:
			writemsg('!!! %s link failed: %s\n' % (p1, e), noiselevel=-1)
			continue
		try:
			os.rename(tmp2, p2)
		except OSError as e:
			writemsg('!!! %s rename failed: %s\n' % (p2, e), noiselevel=-1)
			try:
				os.unlink(tmp2)
			except OSError:
				pass
			continue
		writemsg_stdout('hardlink: %s\n' % filename)

	return os.EX_OK

def usage():
	writemsg('usage: %s <dir1> <dir2>\n' % os.path.basename(sys.argv[0]),
		noiselevel=-1)

def main():
	if len(sys.argv) != 3:
		usage()
		return 1
	return make_hardlinks(*sys.argv[1:])

if __name__ == '__main__':
	sys.exit(main())
